/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

int			playertextures;		// up to 16 color translated skins

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];
float	r_base_world_matrix[16];

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_fullbright = {"r_fullbright","0"};
cvar_t	r_lightmap = {"r_lightmap","0"};
cvar_t	r_shadows = {"r_shadows","0", true};
cvar_t	r_wateralpha = {"r_wateralpha","1", true};
cvar_t	r_dynamic = {"r_dynamic","1"};
cvar_t	r_novis = {"r_novis","0"};

cvar_t	gl_finish = {"gl_finish","0"};
cvar_t	gl_clear = {"gl_clear","0", true};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_keeptjunctions = {"gl_keeptjunctions","0"};
cvar_t	gl_doubleeyes = {"gl_doubleeys", "1"};

// Tomaz - Model Interpolation Begin
cvar_t	r_int_model_anim = { "r_int_model_anim", "1", true };
cvar_t	r_int_model_trans = { "r_int_model_trans", "1", true };
// Tomaz - Model Interpolation End

// Tomaz - Fog Begin
cvar_t  gl_fogenable = {"gl_fogenable", "0", true}; 
cvar_t  gl_fogstart = {"gl_fogstart", "50.0", true}; 
cvar_t  gl_fogend = {"gl_fogend", "800.0", true}; 
cvar_t  gl_fogred = {"gl_fogred","0.6", true};
cvar_t  gl_foggreen = {"gl_foggreen","0.5", true}; 
cvar_t  gl_fogblue = {"gl_fogblue","0.4", true}; 
// Tomaz - Fog End

cvar_t con_alpha = {"con_alpha", "0.5", true};	// Tomaz - Console Alpha
cvar_t r_particles = {"r_particles", "1", true};// Tomaz - New Particle System
cvar_t r_wave = {"r_wave", "0", true};			// Tomaz - Water Wave
cvar_t gl_glows = {"gl_glows", "1", true};		// Tomaz - Glow
cvar_t r_bobbing = {"r_bobbing", "1", true};	// Tomaz - Bobbing Items
cvar_t gl_envmap = {"gl_envmap", "0", true};	// Tomaz - Enviroment Mapping
cvar_t gl_caustics = {"gl_caustics", "0", true};// Tomaz - Underwater Caustics
cvar_t gl_fbr = {"gl_fbr", "1", true};			// Tomaz- Fullbrights
/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	if ((BoxOnPlaneSide (mins, maxs, &frustum[0]) == 2) ||
		(BoxOnPlaneSide (mins, maxs, &frustum[1]) == 2) ||
		(BoxOnPlaneSide (mins, maxs, &frustum[2]) == 2) ||
		(BoxOnPlaneSide (mins, maxs, &frustum[3]) == 2))
		return true;
	return false;
}


void R_RotateForEntity (entity_t *e, int shadow)
{
    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);

	// Tomaz - New Shadow Begin
	if (shadow == 0)
	{
		glRotatef (-e->angles[0],  0, 1, 0);
		glRotatef (e->angles[2],  1, 0, 0);
	}
	// Tomaz - New Shadow End

	glScalef  (e->scale, e->scale, e->scale);	// Tomaz - QC Scale
}

// Tomaz - Model Transform Interpolation Begin

/*
=============
R_BlendedRotateForEntity
=============
*/
void R_BlendedRotateForEntity (entity_t *e, int shadow)	// Tomaz - New Shadow
{
	float timepassed;
	float blend;
	vec3_t d;
	int i;

	// positional interpolation

	timepassed = realtime - e->translate_start_time; 

	if (e->translate_start_time == 0 || timepassed > 1)
	{
		e->translate_start_time = realtime;
		VectorCopy (e->origin, e->origin1);
		VectorCopy (e->origin, e->origin2);
	}

	if (!VectorCompare (e->origin, e->origin2))
	{
		e->translate_start_time = realtime;
		VectorCopy (e->origin2, e->origin1);
		VectorCopy (e->origin,  e->origin2);
		blend = 0;
	}
	else
	{
		blend =  (timepassed / 0.1)*slowmo.value;

		if (cl.paused || blend > 1) blend = 1;
	}

	VectorSubtract (e->origin2, e->origin1, d);

	glTranslatef (
		e->origin1[0] + (blend * d[0]),
		e->origin1[1] + (blend * d[1]),
		e->origin1[2] + (blend * d[2]));

	// orientation interpolation (Euler angles, yuck!)

	timepassed = realtime - e->rotate_start_time; 

	if (e->rotate_start_time == 0 || timepassed > 1)
	{
		e->rotate_start_time = realtime;
		VectorCopy (e->angles, e->angles1);
		VectorCopy (e->angles, e->angles2);
	}

	if (!VectorCompare (e->angles, e->angles2))
	{
		e->rotate_start_time = realtime;
		VectorCopy (e->angles2, e->angles1);
		VectorCopy (e->angles,  e->angles2);
		blend = 0;
	}
	else
	{
		blend = (timepassed / 0.1)*slowmo.value;
 
		if (cl.paused || blend > 1) blend = 1;
	}

	VectorSubtract (e->angles2, e->angles1, d);

	// always interpolate along the shortest path
	for (i = 0; i < 3; i++) 
	{
		if (d[i] > 180)
		{
			d[i] -= 360;
		}
		else if (d[i] < -180)
		{
			d[i] += 360;
		}
	}

	// Tomaz - New Shadow Begin
	glRotatef ( e->angles1[1] + ( blend * d[1]),  0, 0, 1);

	if (shadow==0)
	{
              glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0);
              glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0);
	}
	// Tomaz - New Shadow End

	glScalef  (e->scale, e->scale, e->scale);	// Tomaz - QC Scale
}
// Tomaz - Model Transform Interpolation End

/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

	// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
	// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}

extern qboolean mtexenabled;

/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	// Tomaz - TGA Begin
	glEnable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDepthMask(false);
	// Tomaz - TGA End

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{	// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{	// normal sprite
		up = vup;
		right = vright;
	}

	glColor4f (1,1,1,currententity->alpha);

	if (mtexenabled) 
	{
		glDisable(GL_TEXTURE_2D);
		qglSelectTextureSGIS(TEXTURE0_SGIS);
		mtexenabled = false;
	}

    GL_Bind(frame->gl_texturenum);

	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);
	
	glEnd ();

	// Tomaz - TGA Begin
	glDisable (GL_BLEND);
	glDepthMask(true);
	// Tomaz - TGA End
	glColor4f (1,1,1,1);
}

/*
=============================================================

  ALIAS MODELS

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

vec3_t	shadevector;
// Tomaz - Lit Support

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots = r_avertexnormal_dots[0];

int		lastposenum;
int		lastposenum0;	// Tomaz - Model Animation Interpolation

/*
=============
GL_DrawAliasFrame
=============
*/
extern	vec3_t lightcolor; // Tomaz - Lit Support
void GL_DrawAliasFrame (int frame, aliashdr_t *paliashdr)
{
	float		l, interval;
	int			*order, count, pose, numposes;
	trivertx_t	*verts;

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	lastposenum = pose;

	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += pose * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);
		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			// normals and vertexes come from the frame list
			// Tomaz - Lit Support Begin
			l = shadedots[verts->lightnormalindex];
			glColor4f (l * lightcolor[0], l * lightcolor[1], l * lightcolor[2], currententity->alpha);	// Tomaz - QC Alpha
			// Tomaz - Lit Support End
			glVertex3f (verts->v[0], verts->v[1], verts->v[2]);
			verts++;
		} while (--count);
		glEnd ();
	}
}

// Tomaz - Model Animation Interpolation Begin
/*
=============
GL_DrawAliasBlendedFrame
=============
*/
void GL_DrawAliasBlendedFrame (int frame, aliashdr_t *paliashdr, entity_t* e)
{
	float		l, blend;
	trivertx_t	*verts1, *verts2;
	int			count, pose, numposes, *order;
	vec3_t		d;

	// Tomaz - TGA Begin
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_BLEND);
	// Tomaz - TGA End

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("GL_DrawAliasBlendedFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose		= paliashdr->frames[frame].firstpose;
	numposes	= paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		e->frame_interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / e->frame_interval) % numposes;
	}
	else 
	{
		e->frame_interval = 0.1f;
	}

	if (e->pose2 != pose)
	{
		e->frame_start_time = realtime;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else
	{
		blend = ((realtime - e->frame_start_time)*slowmo.value) / e->frame_interval;

		if (cl.paused || blend > 1) 
			blend = 1;
	}

	lastposenum0 = e->pose1;
	lastposenum  = e->pose2;

	verts1  = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2  = verts1;

	verts1 += e->pose1 * paliashdr->poseverts;
	verts2 += e->pose2 * paliashdr->poseverts;

	order = (int *)((byte *)paliashdr + paliashdr->commands);
 
	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;
        
		if (!count) 
			break;
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
		{
			glBegin (GL_TRIANGLE_STRIP);
		}
		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			// normals and vertexes come from the frame list
			// blend the light intensity from the two frames together
			d[0] = shadedots[verts2->lightnormalindex] - shadedots[verts1->lightnormalindex];

			l = shadedots[verts1->lightnormalindex] + (blend * d[0]);
			glColor4f (l * lightcolor[0], l * lightcolor[1], l * lightcolor[2], currententity->alpha);	// Tomaz - QC Alpha

			VectorSubtract(verts2->v, verts1->v, d);

			// blend the vertex positions from each frame together
			glVertex3f (
						verts1->v[0] + (blend * d[0]),
						verts1->v[1] + (blend * d[1]),
						verts1->v[2] + (blend * d[2]));

			verts1++;
			verts2++;
		} 
		while 
			(--count);
		glEnd ();
	}
	// Tomaz - TGA Begin
	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	// Tomaz - TGA End
}
// Tomaz - Model Animation Interpolation End

/*
=============
GL_DrawAliasShadow
=============
*/
extern	vec3_t			lightspot;

void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
	trivertx_t	*verts;
	int		*order;
	vec3_t	point;
	float	height, lheight;
	int		count;

	// Tomaz - New Shadow Begin
	trace_t		downtrace;
	vec3_t		downmove;
	float		s1,c1;
	// Tomaz - New Shadow End

	lheight = currententity->origin[2] - lightspot[2];

	height = 0;
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	height = -lheight;	// Tomaz - New Shadow

	// Tomaz - New Shadow Begin
	VectorCopy (currententity->origin, downmove);
	downmove[2] = downmove[2] - 4096;
	memset (&downtrace, 0, sizeof(downtrace));
	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, currententity->origin, downmove, &downtrace);

	s1 = sin( currententity->angles[1]/180*M_PI);
	c1 = cos( currententity->angles[1]/180*M_PI);
	// Tomaz - New Shadow End

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			// (skipped for shadows) glTexCoord2fv ((float *)order);
			order += 2;

			// normals and vertexes come from the frame list
			point[0] = verts->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point[1] = verts->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point[2] = verts->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			// Tomaz - New Shadow Begin
			point[0] -= shadevector[0]*(point[0]); 
			point[1] -= shadevector[1]*(point[1]);
			point[2] -= shadevector[2]*(point[2]);

			point[2] = point[2] - (currententity->origin[2] - downtrace.endpos[2]) ;

			point[2] +=	((point[1] * (s1 * downtrace.plane.normal[0])) -
						 (point[0] * (c1 * downtrace.plane.normal[0])) -
						 (point[0] * (s1 * downtrace.plane.normal[1])) - 
						 (point[1] * (c1 * downtrace.plane.normal[1]))) +  
						 ((1.0 - downtrace.plane.normal[2])*20) + 0.2 ;
			// Tomaz - New Shadow End

			glVertex3fv (point);

			verts++;
		} while (--count);

		glEnd ();
	}	
}

// Tomaz - Model Shadow Interpolation Begin
/*
=============
GL_DrawAliasBlendedShadow
=============
*/
void GL_DrawAliasBlendedShadow (aliashdr_t *paliashdr, int pose1, int pose2, entity_t* e)
{
	trivertx_t* verts1;
	trivertx_t* verts2;
	int*        order;
	vec3_t      point1;
	vec3_t      point2;
	vec3_t      d;
	float       height;
	float       lheight;
	int         count;
	float       blend;

	// Tomaz - New Shadow Begin
	trace_t		downtrace;
	vec3_t		downmove;
	float		s1,c1;
	// Tomaz - New Shadow End

	blend = (realtime - e->frame_start_time) / e->frame_interval;

	if (blend > 1) blend = 1;

	lheight = e->origin[2] - lightspot[2];
	height  = -lheight; // Tomaz - New Shadow

	// Tomaz - New Shadow Begin
	VectorCopy (currententity->origin, downmove);
	downmove[2] = downmove[2] - 4096;
	memset (&downtrace, 0, sizeof(downtrace));
	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, currententity->origin, downmove, &downtrace);

	s1 = sin( currententity->angles[1]/180*M_PI);
	c1 = cos( currententity->angles[1]/180*M_PI);
	// Tomaz - New Shadow End

	verts1 = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2 = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *)((byte *)paliashdr + paliashdr->commands);

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;

		if (!count) break;

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
		{
			glBegin (GL_TRIANGLE_STRIP);
		}

		do
		{
			order += 2;

			point1[0] = verts1->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point1[1] = verts1->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point1[2] = verts1->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];
        
			point1[0] -= shadevector[0]*(point1[2]+lheight);
			point1[1] -= shadevector[1]*(point1[2]+lheight);

			point2[0] = verts2->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point2[1] = verts2->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point2[2] = verts2->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			point2[0] -= shadevector[0]*(point2[2]+lheight);
			point2[1] -= shadevector[1]*(point2[2]+lheight);

			VectorSubtract(point2, point1, d);

			// Tomaz - New shadow Begin
			point1[0] = point1[0] + (blend * d[0]);
			point1[1] = point1[1] + (blend * d[1]);
			point1[2] = point1[2] + (blend * d[2]);

			point1[2] =  - (currententity->origin[2] - downtrace.endpos[2]) ;

			point1[2] += ((point1[1] * (s1 * downtrace.plane.normal[0])) -
						  (point1[0] * (c1 * downtrace.plane.normal[0])) -
						  (point1[0] * (s1 * downtrace.plane.normal[1])) - 
						  (point1[1] * (c1 * downtrace.plane.normal[1]))) +  
						  ((1.0 - downtrace.plane.normal[2])*20) + 0.2 ;

			glVertex3fv (point1);
			// Tomaz - New shadow Begin

			verts1++;
			verts2++;
		} 
		while 
			(--count);
		glEnd ();
	}       
}
// Tomaz - Model Shadow Interpolation End

// Tomaz - Quake2 Models Begin
/*
=============
GL_DrawQ2AliasFrame
=============
*/
void GL_DrawQ2AliasFrame (entity_t *e, md2_t *pheader, int lastpose, int pose, float lerp)
{
	float	ilerp;
	float 	l;
	int		*order, count;
	md2trivertx_t	*verts1, *verts2;
	vec3_t	scale1, translate1, scale2, translate2;
	md2frame_t *frame1, *frame2;

	glEnable(GL_BLEND);
	glDepthMask(false);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	ilerp = 1.0 - lerp;

	frame1 = (md2frame_t *)((int) pheader + pheader->ofs_frames + (pheader->framesize * lastpose)); 
	frame2 = (md2frame_t *)((int) pheader + pheader->ofs_frames + (pheader->framesize * pose)); 

	VectorCopy(frame1->scale, scale1);
	VectorCopy(frame1->translate, translate1);
	VectorCopy(frame2->scale, scale2);
	VectorCopy(frame2->translate, translate2);
	verts1 = &frame1->verts[0];
	verts2 = &frame2->verts[0];
	order = (int *)((int)pheader + pheader->ofs_glcmds);

	l = shadedots[verts1->lightnormalindex];
	glColor4f (l * lightcolor[0], l * lightcolor[1], l * lightcolor[2],1);

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			glTexCoord2f(((float *)order)[0], ((float *)order)[1]);
			glVertex3f((verts1[order[2]].v[0]*scale1[0]+translate1[0])*ilerp+(verts2[order[2]].v[0]*scale2[0]+translate2[0])*lerp,
					   (verts1[order[2]].v[1]*scale1[1]+translate1[1])*ilerp+(verts2[order[2]].v[1]*scale2[1]+translate2[1])*lerp,
					   (verts1[order[2]].v[2]*scale1[2]+translate1[2])*ilerp+(verts2[order[2]].v[2]*scale2[2]+translate2[2])*lerp);
				
			order+=3;
		} while (--count);

		glEnd ();
	}

	glDisable(GL_BLEND);
	glDepthMask(true);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}


/*
=============
GL_DrawQ2AliasShadow
=============
*/
void GL_DrawQ2AliasShadow (entity_t *e, md2_t *pheader, int lastpose, int pose, float lerp)
{
	float	ilerp, height, lheight;
	int		*order, count;
	md2trivertx_t	*verts1, *verts2;
	vec3_t	scale1, translate1, scale2, translate2, point;
	md2frame_t *frame1, *frame2;

	lheight = currententity->origin[2] - lightspot[2];

	height = 0;

	ilerp = 1.0 - lerp;

	frame1 = (md2frame_t *)((int) pheader + pheader->ofs_frames + (pheader->framesize * lastpose)); 
	frame2 = (md2frame_t *)((int) pheader + pheader->ofs_frames + (pheader->framesize * pose)); 

	VectorCopy(frame1->scale, scale1);
	VectorCopy(frame1->translate, translate1);
	VectorCopy(frame2->scale, scale2);
	VectorCopy(frame2->translate, translate2);
	verts1 = &frame1->verts[0];
	verts2 = &frame2->verts[0];
	order = (int *)((int) pheader + pheader->ofs_glcmds);

	height = -lheight + 1.0;

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			point[0] = (verts1[order[2]].v[0]*scale1[0]+translate1[0])*ilerp+(verts2[order[2]].v[0]*scale2[0]+translate2[0])*lerp;
			point[1] = (verts1[order[2]].v[1]*scale1[1]+translate1[1])*ilerp+(verts2[order[2]].v[1]*scale2[1]+translate2[1])*lerp;
			point[2] = (verts1[order[2]].v[2]*scale1[2]+translate1[2])*ilerp+(verts2[order[2]].v[2]*scale2[2]+translate2[2])*lerp;

			point[0] -= shadevector[0]*(point[2]+lheight);
			point[1] -= shadevector[1]*(point[2]+lheight);
			point[2] = height;
//			height -= 0.001;
			glVertex3fv(point);
			order+=3;
		} while (--count);

		glEnd ();
	}

}



/*
=================
R_SetupQ2AliasFrame

=================
*/
void R_SetupQ2AliasFrame (entity_t *e, md2_t *pheader)
{
	int				frame;
	float			lerp;

	frame = e->frame;

      glPushMatrix ();
	R_RotateForEntity (e, 0);

	if ((frame >= pheader->num_frames) || (frame < 0))
	{
		Con_DPrintf ("R_SetupQ2AliasFrame: no such frame %d\n", frame);
		frame = 0;
	}

	if (e->draw_lastmodel == e->model)
	{
		if (frame != e->draw_pose)
		{
			e->draw_lastpose = e->draw_pose;
			e->draw_pose = frame;
			e->draw_lerpstart = cl.time;
			lerp = 0;
		}
		else
			lerp = (cl.time - e->draw_lerpstart) * 20.0;
	}
	else // uninitialized
	{
		e->draw_lastmodel = e->model;
		e->draw_lastpose = e->draw_pose = frame;
		e->draw_lerpstart = cl.time;
		lerp = 0;
	}
	if (lerp > 1) lerp = 1;

	GL_DrawQ2AliasFrame (e, pheader, e->draw_lastpose, frame, lerp);

	if (r_shadows.value)
	{
		glDisable (GL_TEXTURE_2D);
		glEnable (GL_BLEND);
		glDepthMask(false); // disable zbuffer updates
		glColor4f (0,0,0,0.5);// * modelalpha);
		GL_DrawQ2AliasShadow (e, pheader, e->draw_lastpose, frame, lerp);
		glDepthMask(true); // enable zbuffer updates
		glEnable (GL_TEXTURE_2D);
//		glDisable (GL_BLEND);
		glColor3f (1,1,1);
	}
	glPopMatrix ();
}
// Tomaz - Quake2 Models End

/*
=================
R_DrawAliasModel
=================
*/
void R_DrawAliasModel (entity_t *e)
{
	int			i, j, lnum, anim;
	float		add, an;
	md2_t		*pheader; // Tomaz - Quake2 Models
	vec3_t		dist, mins, maxs;
	model_t		*clmodel;
	aliashdr_t	*paliashdr;

	// Tomaz - New Shadow Begin
	trace_t		downtrace;
	vec3_t		downmove;
	// Tomaz - New Shadow End

	clmodel = currententity->model;

	VectorAdd (currententity->origin, clmodel->mins, mins);
	VectorAdd (currententity->origin, clmodel->maxs, maxs);

	if (R_CullBox (mins, maxs))
		return;

	VectorSubtract (r_origin, currententity->origin, modelorg);

	//
	// get lighting information
	//

	R_LightPoint(currententity->origin); // Tomaz - Lit Support

	// allways give the gun some light

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (currententity->origin,
							cl_dlights[lnum].origin,
							dist);
			add = cl_dlights[lnum].radius - Length(dist);

			// Tomaz - Lit Support Begin
			if (add > 0)
			{
				lightcolor[0] += add * cl_dlights[lnum].color[0];
				lightcolor[1] += add * cl_dlights[lnum].color[1];
				lightcolor[2] += add * cl_dlights[lnum].color[2];
			}
			// Tomaz - Lit Support End
		}
	}

	// Tomaz - Lit Support

	// ZOID: never allow players to go totally black
	i = currententity - cl_entities;

	// HACK HACK HACK -- no fullbright colors, so make torches full light
	if (!strcmp (clmodel->name, "progs/flame2.mdl")	|| 
		!strcmp (clmodel->name, "progs/flame.mdl") ||
		!strncmp(clmodel->name, "progs/bolt", 10))
		lightcolor[0] = lightcolor[1] = lightcolor[2] = 256;	// Tomaz - Lit Support

	shadedots = r_avertexnormal_dots[((int)(e->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];

	// Tomaz - Lit Support Begin
	lightcolor[0] = lightcolor[0] / 200.0f;
	lightcolor[1] = lightcolor[1] / 200.0f;
	lightcolor[2] = lightcolor[2] / 200.0f;
	// Tomaz - Lit Support End

	an = e->angles[1]/180*M_PI;
	shadevector[0] = cos(-an);
	shadevector[1] = sin(-an);
	shadevector[2] = 1;
	VectorNormalize (shadevector);

	//
	// locate the proper data
	//

	// Tomaz - Quake2 Models Begin
	if (clmodel->aliastype == ALIASTYPE_MD2)
	{
		pheader = (md2_t *)Mod_Extradata (currententity->model);
		c_alias_polys += pheader->num_tris;
	}
	else
	{
		paliashdr = (aliashdr_t *)Mod_Extradata (currententity->model);
		c_alias_polys += paliashdr->numtris;
	}
	// Tomaz - Quake2 Models End

	//
	// draw all the triangles
	//

	if (mtexenabled) 
	{
		glDisable(GL_TEXTURE_2D);
		qglSelectTextureSGIS(TEXTURE0_SGIS);
		mtexenabled = false;
	}

	// Tomaz - Quake2 Models Begin
	if (clmodel->aliastype != ALIASTYPE_MD2)
	{
		glPushMatrix ();
		// Tomaz - Model Transform Interpolation Begin
		if (r_int_model_trans.value)
		{
			R_BlendedRotateForEntity (e, 0);	// Tomaz - New Shadow
		}
		else
		{
			R_RotateForEntity (e, 0);	// Tomaz - New Shadow
		}
		// Tomaz - Model Transform Interpolation End

		if (!strcmp (clmodel->name, "progs/eyes.mdl") && gl_doubleeyes.value) 
		{
			glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2] - (22 + 8));
// double size of eyes, since they are really hard to see in gl
			glScalef (paliashdr->scale[0]*2, paliashdr->scale[1]*2, paliashdr->scale[2]*2);
		}
		else
		{
			glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
			glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);
		}
	}
	// Tomaz - Quake2 Models End

	// Tomaz - Quake2 Models Begin
	if (clmodel->aliastype == ALIASTYPE_MD2)
	    GL_Bind(pheader->gl_texturenum[currententity->skinnum]);
	else
	{
		anim = (int)(cl.time*10) & 3;
	    GL_Bind(paliashdr->gl_texturenum[currententity->skinnum][anim]);
	}
	// Tomaz - Quake2 Models End

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (currententity->colormap != vid.colormap && !gl_nocolors.value)
	{
		i = currententity - cl_entities;
		if (i >= 1 && i<=cl.maxclients /* && !strcmp (currententity->model->name, "progs/player.mdl") */)
		    GL_Bind(playertextures - 1 + i);
	}

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	// Tomaz - Quake2 Models Begin
	if (clmodel->aliastype == ALIASTYPE_MD2)
		R_SetupQ2AliasFrame (currententity, pheader);
	else
	{
		// Tomaz - Model Animation Interpolation Begin
		if (r_int_model_anim.value)
		{
			GL_DrawAliasBlendedFrame (currententity->frame, paliashdr, currententity);
		}
		else
		{
			GL_DrawAliasFrame (currententity->frame, paliashdr);
		}
		// Tomaz - Model Animation Interpolation End
	}
	// Tomaz - Quake2 Models End

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	// Tomaz - Quake2 Models Begin
	if (clmodel->aliastype != ALIASTYPE_MD2)
	{
		glPopMatrix ();

		// Tomaz - Glow Begin
		if  (gl_glows.value)
		{
			if ( (!strncmp (clmodel->name, "progs/flame",11)) ||
				 (!strncmp (clmodel->name, "progs/glow_",11)) ||
				 (!strncmp (clmodel->name, "progs/bolt",10)) ||
				 (!strcmp (clmodel->name, "progs/plasma.mdl")) ||
				 (!strcmp (clmodel->name, "progs/missile.mdl")) ||
				 (!strcmp (clmodel->name, "progs/laser.mdl")) )
			{
				vec3_t	lightorigin;	// Origin of torch.
				vec3_t	v;				// Vector to torch.
				float	radius;			// Radius of torch flare.
				float	distance;		// Vector distance to torch.
				float	intensity;		// Intensity of torch flare.	
	
				// NOTE: I don't think this is centered on the model.
				VectorCopy(currententity->origin, lightorigin);	

				//set radius based on what model we are doing here
				if ((!strcmp (clmodel->name, "progs/missile.mdl")) ||
					(!strcmp (clmodel->name, "progs/plasma.mdl")))
					radius = 6.0f;
				else if( (!strncmp (clmodel->name, "progs/flame",11)) ||
						 (!strncmp (clmodel->name, "progs/glow_",11)) ||
						 (!strncmp (clmodel->name, "progs/bolt",10))      ||
						 (!strcmp (clmodel->name, "progs/laser.mdl")) )
					radius = 24.0f;
	
				VectorSubtract(lightorigin, r_origin, v);

				// See if view is outside the light.
				distance = Length(v);
				if (distance > radius) 
				{
					glDepthMask (false);
					glDisable (GL_TEXTURE_2D);
					glEnable (GL_BLEND);
					glBlendFunc (GL_ONE, GL_ONE);

					glPushMatrix();			

					glBegin(GL_TRIANGLE_FAN);

					if ((!strncmp (clmodel->name, "progs/flame",11)) || 
						(!strncmp (clmodel->name, "progs/glow_",11)))
					{	
						glTranslatef(0.0f, 0.0f, 4.0f);
					    intensity = (1024.0f - distance) / 1024.0f;
						intensity = ((1.0f - intensity)*0.5);		
						glColor3f(1.0f*intensity, 0.7f*intensity, 0.4f*intensity);		
					}
					else if (!strcmp (clmodel->name, "progs/missile.mdl"))
					{
						intensity = 0.7f;
						glColor3f(1.0f*intensity, 0.7f*intensity, 0.4f*intensity);
					}
					else if (!strcmp (clmodel->name, "progs/plasma.mdl"))
					{
						intensity = 0.7f;
						glColor3f(0.0f*intensity, 1.0f*intensity, 0.0f*intensity);
					}
					else if (!strcmp (clmodel->name, "progs/bolt.mdl"))
					{
						intensity = 0.2f;
						glColor3f(1.0f*intensity, 0.3f*intensity, 0.3f*intensity);
					}
					else if (!strcmp (clmodel->name, "progs/bolt2.mdl"))
					{
						intensity = 0.2f;
						glColor3f(0.3f*intensity, 0.3f*intensity, 1.0f*intensity);				
					}
					else if (!strcmp (clmodel->name, "progs/bolt3.mdl"))
					{
						intensity = 0.2f;
						glColor3f(0.3f*intensity, 0.3f*intensity, 1.0f*intensity);				
					}
					else if (!strcmp (clmodel->name, "progs/laser.mdl"))
					{
						intensity = 0.2f;			
						glColor3f(1.0f*intensity, 0.3f*intensity, 0.3f*intensity);
					}
					// Clamp, but don't let the flare disappear.
					if (intensity > 1.0f) intensity = 1.0f;
					if (intensity < 0.0f) intensity = 0.0f;
		
					for (i=0 ; i<3 ; i++)
						v[i] = lightorigin[i] - vpn[i]*radius;
					glVertex3fv(v);
					glColor3f(0.0f, 0.0f, 0.0f);
					for (i=16; i>=0; i--) 
					{
						float a = i/16.0f * M_PI*2;
						for (j=0; j<3; j++)
							v[j] =	lightorigin[j] + 
									vright[j]*cos(a)*radius +
									vup[j]*sin(a)*radius;
						glVertex3fv(v);
					}
					glEnd();
					glPopMatrix();
					glColor3f (1,1,1);
					glDisable (GL_BLEND);
					glEnable (GL_TEXTURE_2D);
					glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					glDepthMask (true);
				}
			}
		}	
		// Tomaz - Glow End

		// Tomaz - New Shadow Begin
		if (r_shadows.value)
		{

			if ((!strncmp (clmodel->name, "progs/v_",8)) ||
				(!strncmp (clmodel->name, "progs/bolt",10)) ||
				(!strncmp (clmodel->name, "progs/flame",11)))
			return;

			glPushMatrix ();

			shadevector[0] = 0;
			shadevector[1] = 0;
			shadevector[2] = 1;
			VectorNormalize (shadevector);

			// Tomaz - Model Shadow Interpolation Begin
			if (r_int_model_trans.value)
			{
				R_BlendedRotateForEntity (e, 1);
			}
			else
			{
				R_RotateForEntity (e, 1);
			}
			// Tomaz - Model Shadow Interpolation End

			VectorCopy (e->origin, downmove);

			downmove[2] = downmove[2] - 4096;
			memset (&downtrace, 0, sizeof(downtrace));
			SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, e->origin, downmove, &downtrace);

			glDisable (GL_TEXTURE_2D);
			glEnable (GL_BLEND);
			glColor4f (0,0,0,1.2 - ((mins[2]-downtrace.endpos[2])/60));

			// Tomaz - Model Shadow Interpolation Begin
			if (r_int_model_anim.value)	
			{	
				GL_DrawAliasBlendedShadow (paliashdr, lastposenum0, lastposenum, currententity);
			}
			else
			{
				GL_DrawAliasShadow (paliashdr, lastposenum);
			}
			// Tomaz - Model Shadow Interpolation End
			glEnable (GL_TEXTURE_2D);
			glDisable (GL_BLEND);
			glColor4f (1,1,1,1);
			glPopMatrix ();
		}
		// Tomaz - New Shadow End
	}
	// Tomaz - Quake2 Models End

	// Tomaz - TGA Begin
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	// Tomaz - TGA End
}
//==================================================================================

/*
=============
R_DrawBrushEntities
=============
*/
void R_DrawBrushEntities (void)
{
	int		i;

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		if (currententity->model->type != mod_brush)
			continue;

		if ((currententity->alpha != 1) && (currententity->alpha != 0))
		{
			currententity->solid = false;
			continue;
		}

		R_DrawBrushModel (currententity);
	}
}

/*
=============
R_DrawModelEntities
=============
*/
void R_DrawModelEntities (void)
{
	int		i;

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];
		
		if (currententity->model->type != mod_alias)
			continue;

		if ((currententity->alpha != 1) && (currententity->alpha != 0))
		{
			currententity->solid = false;
			continue;
		}

		// Tomaz - Chase Cam Fix Begin
		if (currententity == &cl_entities[cl.viewentity]) 
			currententity->angles[0] *= 0.3f;
		// Tomaz - Chase Cam Fix End

		// Tomaz - QC Alpha Scale Begin
		if ((!strncmp (currententity->model->name, "progs/bolt", 10)) ||
			(!strncmp (currententity->model->name, "progs/flame", 11)))
		{
			currententity->alpha = 0.5;
			currententity->scale = 1;
		}
		// Tomaz - QC Alpha Scale End

		R_DrawAliasModel (currententity);
	}
}

/*
=============
R_DrawSpriteEntities
=============
*/
void R_DrawSpriteEntities (void)
{
	int		i;

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		if (currententity->model->type != mod_sprite)
			continue;

		R_DrawSpriteModel (currententity);
	}
}

/*
=============
R_DrawTransBrushes
=============
*/
void R_DrawTransBrushes (void)
{

	int			i;
	float		bestdist, dist;
	entity_t	*bestent;
	vec3_t		start, test;
	
	VectorCopy(r_refdef.vieworg, start);

transgetent:
	bestdist = 0;
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		if ((currententity->solid) ||
			(currententity->alpha == 1) ||
			(currententity->model->type != mod_brush))
			continue;

		VectorCopy(currententity->origin, test);

		test[0] += currententity->model->mins[0];
		test[1] += currententity->model->mins[1];
		test[2] += currententity->model->mins[2];

		dist = (((test[0] - start[0]) * (test[0] - start[0])) +
				((test[1] - start[1]) * (test[1] - start[1])) +
				((test[2] - start[2]) * (test[2] - start[2])));

		if (dist > bestdist)
		{
			bestdist = dist;
			bestent = currententity;
		
		}
	}

	if (bestdist == 0)
		return;

	bestent->solid = true;

	currententity = bestent;

	R_DrawBrushModel (currententity);
	
	goto transgetent;
}

/*
=============
R_DrawTransModels
=============
*/
void R_DrawTransModels (void)
{
	int			i;
	float		bestdist, dist;
	entity_t	*bestent;
	vec3_t		start, test;
	
	VectorCopy(r_refdef.vieworg, start);

transgetent:
	bestdist = 0;
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		if ((currententity->solid) ||
			(currententity->alpha == 1) ||
			(currententity->model->type != mod_alias))
			continue;

		VectorCopy(currententity->origin, test);

		dist = (((test[0] - start[0]) * (test[0] - start[0])) +
				((test[1] - start[1]) * (test[1] - start[1])) +
				((test[2] - start[2]) * (test[2] - start[2])));

		if (dist > bestdist)
		{
			bestdist	= dist;
			bestent		= currententity;
		
		}
	}

	if (bestdist == 0)
		return;

	bestent->solid = true;

	currententity = bestent;

	R_DrawAliasModel (currententity);
	
	goto transgetent;
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{

	float old_model;	// Tomaz - Model Transform Interpolation

	// Tomaz - Lit Support

	if (!r_drawviewmodel.value)
		return;

	if (chase_active.value)
		return;

	if (cl.items & IT_INVISIBILITY)
		return;

	if (cl.stats[STAT_HEALTH] <= 0)
		return;

	currententity = &cl.viewent;
	if (!currententity->model)
		return;

	// Tomaz - QC Alpha Scale Begin
	currententity->alpha = cl_entities[cl.viewentity].alpha;
	currententity->scale = cl_entities[cl.viewentity].scale;
	// Tomaz - QC Alpha Scale End

	// Tomaz - Lit Support

	// hack the depth range to prevent view model from poking into walls
	glDepthRange (gldepthmin, gldepthmin + 0.3*(gldepthmax-gldepthmin));

	// Tomaz - Model Transform Interpolation Begin
	old_model = r_int_model_trans.value;
	r_int_model_trans.value = false;
	R_DrawAliasModel (currententity);
	r_int_model_trans.value = old_model;
	// Tomaz - Model Transform Interpolation Begin
	glDepthRange (gldepthmin, gldepthmax);
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!v_blend[3])
		return;

	if (mtexenabled) 
	{
		glDisable(GL_TEXTURE_2D);
		qglSelectTextureSGIS(TEXTURE0_SGIS);
		mtexenabled = false;
	}

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);
	glEnd ();

	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}


void R_SetFrustum (void)
{
	int		i;

	if (r_refdef.fov_x == 90) 
	{
		// front side is visible

		VectorAdd (vpn, vright, frustum[0].normal);
		VectorSubtract (vpn, vright, frustum[1].normal);

		VectorAdd (vpn, vup, frustum[2].normal);
		VectorSubtract (vpn, vup, frustum[3].normal);
	}
	else
	{
		// rotate VPN right by FOV_X/2 degrees
		RotatePointAroundVector( frustum[0].normal, vup, vpn, -(90-r_refdef.fov_x / 2 ) );
		// rotate VPN left by FOV_X/2 degrees
		RotatePointAroundVector( frustum[1].normal, vup, vpn, 90-r_refdef.fov_x / 2 );
		// rotate VPN up by FOV_X/2 degrees
		RotatePointAroundVector( frustum[2].normal, vright, vpn, 90-r_refdef.fov_y / 2 );
		// rotate VPN down by FOV_X/2 degrees
		RotatePointAroundVector( frustum[3].normal, vright, vpn, -( 90 - r_refdef.fov_y / 2 ) );
	}

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{

// don't allow cheats in multiplayer
	if (cl.maxclients > 1)
		Cvar_Set ("r_fullbright", "0");

	R_AnimateLight ();

	r_framecount++;

// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;

	c_brush_polys = 0;
	c_alias_polys = 0;

}


void MYgluPerspective( GLdouble fovy, GLdouble aspect,
		     GLdouble zNear, GLdouble zFar )
{
   GLdouble xmin, xmax, ymin, ymax;

   ymax = zNear * tan( fovy * 0.008726646259971);	// Tomaz Speed
   ymin = -ymax;

   xmin = ymin * aspect;
   xmax = ymax * aspect;

   glFrustum( xmin, xmax, ymin, ymax, zNear, zFar );
}


/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	float	screenaspect;
	extern	int glwidth, glheight;
	int		x, x2, y2, y, w, h;

	//
	// set up viewpoint
	//
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	x = r_refdef.vrect.x * glwidth/vid.width;
	x2 = (r_refdef.vrect.x + r_refdef.vrect.width) * glwidth/vid.width;
	y = (vid.height-r_refdef.vrect.y) * glheight/vid.height;
	y2 = (vid.height - (r_refdef.vrect.y + r_refdef.vrect.height)) * glheight/vid.height;

	// fudge around because of frac screen scale
	if (x > 0)
		x--;
	if (x2 < glwidth)
		x2++;
	if (y2 < 0)
		y2--;
	if (y < glheight)
		y++;

	w = x2 - x;
	h = y - y2;

	glViewport (glx + x, gly + y2, w, h);
    screenaspect = (float)r_refdef.vrect.width/r_refdef.vrect.height;
    MYgluPerspective (r_refdef.fov_y,  screenaspect,  4,  4096);

	glCullFace(GL_FRONT);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up
    glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
    glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
    glRotatef (-r_refdef.viewangles[1],  0, 0, 1);
    glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	//
	// set drawing parms
	//

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}

/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	if (gl_clear.value)
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	else
		glClear (GL_DEPTH_BUFFER_BIT);
	gldepthmin = 0;
	gldepthmax = 1;

	glDepthRange (gldepthmin, gldepthmax);
}

/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	double	time1, time2;
	GLfloat colors[4] = {(GLfloat) 0.0, (GLfloat) 0.0, (GLfloat) 1, (GLfloat) 0.20};

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	if (r_speeds.value)
	{
		glFinish ();
		time1 = Sys_FloatTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
	}

	if (gl_finish.value)
		glFinish ();

	R_Clear ();

	R_SetupFrame ();

	R_SetFrustum ();

	R_SetupGL ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();		// adds static entities to the list

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	R_DrawBrushEntities ();

	R_DrawModelEntities ();

	R_DrawSpriteEntities ();

	if (mtexenabled) 
	{
		glDisable(GL_TEXTURE_2D);
		qglSelectTextureSGIS(TEXTURE0_SGIS);
		mtexenabled = false;
	}

	R_DrawViewModel ();

	R_DrawWaterSurfaces ();

	R_DrawParticles ();	// Tomaz - Particle System

	R_DrawTransBrushes ();

	R_DrawTransModels ();

	R_PolyBlend ();

	if (r_speeds.value)
	{
//		glFinish ();
		time2 = Sys_FloatTime ();
		Con_Printf ("%3i ms  %4i wpoly %4i epoly\n", (int)((time2-time1)*1000), c_brush_polys, c_alias_polys); 
	}
}
